from dataclasses import dataclass
from typing import cast

import requests

from .models import Board, CardWithID, List
from .settings import TrelloSettings


@dataclass
class BoardShallow:
    id: str
    name: str


type ListID = str


class TrelloClient:
    def __init__(self, board_id: str, api_key: str, api_secret: str):
        self._board_id = board_id
        self._api_key = api_key
        self._api_secret = api_secret
        self._session = requests.Session()
        self._session.params.update({"key": self._api_key, "token": self._api_secret})

    @classmethod
    def from_env(cls) -> "TrelloClient":
        settings = TrelloSettings()  # type: ignore
        return cls(
            board_id=settings.trello_board_id,
            api_key=settings.trello_api_key,
            api_secret=settings.trello_api_secret,
        )

    def list_boards(self) -> list[BoardShallow]:
        url = "https://api.trello.com/1/members/me/boards"
        response = self._session.get(url)
        response.raise_for_status()
        boards_json = response.json()
        return [
            BoardShallow(id=board["id"], name=board["name"]) for board in boards_json
        ]

    def get_board(self) -> Board:
        response_board = self._session.get(
            f"https://api.trello.com/1/boards/{self._board_id}",
            params={"fields": "name"},
        )
        response_board.raise_for_status()
        board_json = response_board.json()

        response_lists = self._session.get(
            f"https://api.trello.com/1/boards/{self._board_id}/lists",
            params={"fields": "name"},
        )
        response_lists.raise_for_status()
        lists_json = response_lists.json()

        response_cards = self._session.get(
            f"https://api.trello.com/1/boards/{self._board_id}/cards",
            params={"fields": "name,idList"},
        )
        response_cards.raise_for_status()
        cards_json = response_cards.json()

        grouped_cards: dict[ListID, list[dict]] = {}
        for card in cards_json:
            grouped_cards.setdefault(card["idList"], []).append(card)

        lists: list[List] = []
        for list_ in lists_json:
            cards_for_this_list = grouped_cards.get(list_["id"], [])
            cards = [
                CardWithID(id=card["id"], name=card["name"])
                for card in cards_for_this_list
            ]
            lists.append(List(id=list_["id"], name=list_["name"], cards=cards))

        return Board(
            id=cast(str, board_json["id"]), name=board_json["name"], lists=lists
        )

    def create_card(self, name: str, list_id: str):
        url = "https://api.trello.com/1/cards"
        response = self._session.post(url, json={"name": name, "idList": list_id})
        response.raise_for_status()
        card_json = response.json()
        return card_json["id"]

    def change_parent(self, card_id: str, new_list_id: str):
        url = f"https://api.trello.com/1/cards/{card_id}"
        response = self._session.put(url, json={"idList": new_list_id})
        response.raise_for_status()

    def rename_card(self, card_id: str, new_name: str):
        url = f"https://api.trello.com/1/cards/{card_id}"
        response = self._session.put(url, json={"name": new_name})
        response.raise_for_status()

    def archive_card(self, card_id: str):
        url = f"https://api.trello.com/1/cards/{card_id}"
        response = self._session.put(url, json={"closed": True})
        response.raise_for_status()


__all__ = ["TrelloClient"]



from rebello.models import Board, CardNoID, List

board = Board(
    id="1",
    name="Board",
    lists=[
        List(
            id="2",
            name="Col",
            cards=[
                CardNoID(name="Say hello"),
            ],
        ),
    ],
)



from pathlib import Path

import pytest
from .edi import temp_path, open_in_editor


class TestTempPath:
    @staticmethod
    def test_sample():
        with temp_path(".md") as path:
            assert path is not None
            assert ".md" in str(path)

            path.write_text("hello")
            assert path.read_text() == "hello"


class TestOpenInEditor:
    @staticmethod
    def test_success(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        tmp_file = tmp_path / "hello.txt"
        tmp_file.write_text("hello")

        # Always returns status code 0.
        monkeypatch.setenv("EDITOR", "true")

        success = open_in_editor(tmp_file)

        assert success is True

    @staticmethod
    def test_failure(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        tmp_file = tmp_path / "hello.txt"
        tmp_file.write_text("hello")

        # Always returns status code 0.
        monkeypatch.setenv("EDITOR", "false")

        success = open_in_editor(tmp_file)

        assert success is False



from pathlib import Path
import re

from .models import Board, List, CardWithID, CardNoID


def render_board(path: Path, board: Board[CardWithID | CardNoID]):
    with path.open("w") as f:
        f.write(f"# [{board.id}] {board.name}\n")

        for col in board.lists:
            f.write("\n")
            f.write(f"## [{col.id}] {col.name}\n")
            f.write("\n")

            for card in col.cards:
                if isinstance(card, CardWithID):
                    f.write(f"* [{card.id}] {card.name}\n")
                else:
                    f.write(f"* {card.name}\n")


def parse_board(path: Path) -> Board[CardWithID | CardNoID]:
    with path.open() as f:
        board_id: str | None = None
        board_name: str | None = None

        lists: list[List] = []

        for line_with_n in f:
            line = line_with_n.strip()
            if (h1_match := re.match(r"^\s*# \[(\w+)\] (.*)$", line)) is not None:
                board_id = h1_match.group(1)
                board_name = h1_match.group(2)

            elif (h2_match := re.match(r"^\s*## \[(\w+)\] (.*)$", line)) is not None:
                # We'll embrace deeply mutable data structures. Normally I don't do it but it's handy here. We create
                # `cards=[]` ahead of time and will mutate it later on.
                lists.append(
                    List(id=h2_match.group(1), name=h2_match.group(2), cards=[])
                )

            elif (li_match := re.match(r"^\s*[*-] \[(\w+)\] (.*)$", line)) is not None:
                try:
                    latest_list = lists[-1]
                except IndexError:
                    # TODO: add error handling
                    pass

                latest_list.cards.append(
                    CardWithID(id=li_match.group(1), name=li_match.group(2))
                )

            elif (li_match := re.match(r"^\s*[*-] (.*)$", line)) is not None:
                try:
                    latest_list = lists[-1]
                except IndexError:
                    # TODO: add error handling
                    pass

                latest_list.cards.append(CardNoID(name=li_match.group(1)))

    if not board_id or not board_name:
        raise ValueError("No valid board header in file", path)

    return Board(id=board_id, name=board_name, lists=lists)



import pytest
from pathlib import Path
import importlib

from .models import Board, CardWithID
from .md import render_board, parse_board


TEST_PKG = "rebello.md_test_files.pairs"
TEST_DIR = Path(__file__).parent / "md_test_files" / "pairs"
EXAMPLE_NAMES = [path.stem for path in TEST_DIR.glob("*.py")]
MALFORMED_DIRS = Path(__file__).parent / "md_test_files" / "malformed"


class TestRenderBoard:
    @staticmethod
    @pytest.mark.parametrize("example_name", EXAMPLE_NAMES)
    def test_module_to_md(example_name: str, tmp_path: Path):
        input_module = importlib.import_module(f"{TEST_PKG}.{example_name}")
        board: Board[CardWithID] = getattr(input_module, "board")
        ref_out = (TEST_DIR / (example_name + ".md")).read_text()
        out_file = tmp_path / "out.md"

        render_board(out_file, board)

        rendered_text = out_file.read_text()
        assert rendered_text == ref_out


class TestParseBoard:
    @staticmethod
    @pytest.mark.parametrize("example_name", EXAMPLE_NAMES)
    def test_md_to_module_contents(example_name: str):
        in_file = TEST_DIR / (example_name + ".md")
        module = importlib.import_module(f"{TEST_PKG}.{example_name}")
        ref_board: Board = getattr(module, "board")

        parsed_board = parse_board(in_file)

        assert parsed_board is not None
        assert parsed_board == ref_board

    @staticmethod
    @pytest.mark.parametrize("example_name", ["empty", "no_board_id"])
    def test_malformed_board(example_name: str):
        in_file = MALFORMED_DIRS / (example_name + ".md")

        with pytest.raises(ValueError) as e:
            _ = parse_board(in_file)

        assert e.match(r"No valid board header")

    @staticmethod
    @pytest.mark.parametrize("example_name", ["no_list_id"])
    def test_malformed_list(example_name: str):
        in_file = MALFORMED_DIRS / (example_name + ".md")

        board = parse_board(in_file)

        assert len(board.lists)



from argparse import ArgumentParser

from .models import ArchiveCard, ChangeList, CreateCard, RewordCard
from .trello import TrelloClient
from .md import render_board, parse_board
from .edi import open_in_editor, temp_path
from .diff import find_changeset


def main():
    parser = ArgumentParser(description="Edits a trello board")
    parser.parse_args()

    client = TrelloClient.from_env()
    board = client.get_board()

    with temp_path(".md") as path:
        render_board(path, board)
        success = open_in_editor(path)
        if not success:
            print(f"Error editing {path}")
            exit(1)

        edited_board = parse_board(path)

    changeset = find_changeset(board, edited_board)
    print(f"Executing {len(changeset)} changes:")
    for change in changeset:
        if isinstance(change, ChangeList):
            print(
                f"* for [{change.card_id}], "
                f"change parent [{change.src_list_id}] -> [{change.dest_list_id}]"
            )
            client.change_parent(
                card_id=change.card_id, new_list_id=change.dest_list_id
            )
        elif isinstance(change, RewordCard):
            print(
                f"* for [{change.card_id}], "
                f"rename [{change.old_name}] -> [{change.new_name}]"
            )
            client.rename_card(card_id=change.card_id, new_name=change.new_name)
        elif isinstance(change, ArchiveCard):
            print(f"* archive [{change.card_id}]")
            client.archive_card(card_id=change.card_id)
        elif isinstance(change, CreateCard):
            print(f"* create '{change.name}' under [{change.list_id}]")
            client.create_card(name=change.name, list_id=change.list_id)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Protocol
from .models import (
    ArchiveCard,
    Board,
    CardNoID,
    CardWithID,
    ChangeList,
    Changeset,
    CreateCard,
    RewordCard,
)


class HasID(Protocol):
    @property
    def id(self) -> str: ...


@dataclass
class CardShallow:
    id: str
    name: str
    list_id: str


def find_changeset(
    board1: Board[CardWithID], board2: Board[CardWithID | CardNoID]
) -> Changeset:
    list1_ids = {a_list.id for a_list in board1.lists}
    list2_ids = {a_list.id for a_list in board2.lists}
    common_list_ids = list1_ids.intersection(list2_ids)

    cards1 = {
        card.id: CardShallow(id=card.id, name=card.name, list_id=a_list.id)
        for a_list in board1.lists
        if a_list.id in common_list_ids
        for card in a_list.cards
    }
    cards2 = {
        card.id: CardShallow(id=card.id, name=card.name, list_id=a_list.id)
        for a_list in board2.lists
        if a_list.id in common_list_ids
        for card in a_list.cards
        if isinstance(card, CardWithID)
    }

    changeset = []

    removed_ids = set(cards1.keys()).difference(cards2.keys())
    for removed_id in removed_ids:
        changeset.append(ArchiveCard(card_id=removed_id))

    common_ids = set(cards1.keys()).intersection(cards2.keys())
    for common_id in common_ids:
        card1 = cards1[common_id]
        card2 = cards2[common_id]

        if card2.name != card1.name:
            changeset.append(
                RewordCard(
                    card_id=common_id,
                    old_name=card1.name,
                    new_name=card2.name,
                )
            )

        if card2.list_id != card1.list_id:
            changeset.append(
                ChangeList(
                    card_id=common_id,
                    src_list_id=card1.list_id,
                    dest_list_id=card2.list_id,
                )
            )

    for a_list in board2.lists:
        for card in a_list.cards:
            if not isinstance(card, CardNoID):
                continue
            changeset.append(CreateCard(list_id=a_list.id, name=card.name))

    return changeset



from .diff import find_changeset
from copy import deepcopy
from .models import (
    ArchiveCard,
    Board,
    CardNoID,
    ChangeList,
    CreateCard,
    List,
    CardWithID,
    RewordCard,
)


BOARD = Board(
    id="a1",
    name="Simple kanban",
    lists=[
        List(
            id="b2",
            name="To Do",
            cards=[
                CardWithID(id="c21", name="Say hello"),
                CardWithID(id="c22", name="Say goodnight"),
            ],
        ),
        List(
            id="b3",
            name="Done",
            cards=[
                CardWithID(id="c31", name="Basic project setup"),
            ],
        ),
    ],
)


class TestFindChangeset:
    @staticmethod
    def test_no_changes():
        assert find_changeset(board1=BOARD, board2=BOARD) == []

    @staticmethod
    def test_missing_list():
        board1 = BOARD
        board2 = deepcopy(board1)
        board2.lists[:] = []

        assert find_changeset(board1, board2) == []

    @staticmethod
    def test_removed_card():
        board1 = BOARD
        board2 = deepcopy(board1)
        archived_card = board2.lists[0].cards.pop(-1)

        assert find_changeset(board1, board2) == [ArchiveCard(card_id=archived_card.id)]

    @staticmethod
    def test_reword():
        board1 = BOARD
        board2 = deepcopy(board1)
        card = board2.lists[0].cards[0]
        old_name = card.name
        card.name = "changed"

        assert find_changeset(board1, board2) == [
            RewordCard(card_id=card.id, old_name=old_name, new_name=card.name)
        ]

    @staticmethod
    def test_creating_card_in_non_empty_list():
        board1: Board[CardNoID | CardWithID] = BOARD
        board2: Board[CardNoID | CardWithID] = deepcopy(board1)

        new_card = CardNoID(name="new name")
        cards = board2.lists[0].cards
        cards.append(new_card)  # type: ignore

        assert find_changeset(board1, board2) == [
            CreateCard(list_id=board2.lists[0].id, name=new_card.name)
        ]

    @staticmethod
    def test_creating_card_in_empty_list():
        board1: Board[CardNoID | CardWithID] = Board(
            id="a1",
            name="a1",
            lists=[List(id="b1", name="b1", cards=[])],
        )
        board2: Board[CardNoID | CardWithID] = deepcopy(board1)

        new_card = CardNoID(name="new name")
        cards = board2.lists[0].cards
        cards.append(new_card)  # type: ignore

        assert find_changeset(board1, board2) == [
            CreateCard(list_id=board2.lists[0].id, name=new_card.name)
        ]

    @staticmethod
    def test_move_card():
        board1: Board[CardNoID | CardWithID] = BOARD
        board2: Board[CardNoID | CardWithID] = deepcopy(board1)

        card = board2.lists[0].cards.pop(-1)
        board2.lists[1].cards.append(card)

        assert find_changeset(board1, board2) == [
            ChangeList(
                card_id=card.id,
                src_list_id=board1.lists[0].id,
                dest_list_id=board2.lists[1].id,
            )
        ]



import json
from pathlib import Path
import re
import pytest
import responses
from .trello import TrelloClient


class TestTrelloClient:
    @staticmethod
    @pytest.fixture
    def client(monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("TRELLO_API_KEY", "api-key")
        monkeypatch.setenv("TRELLO_API_SECRET", "api-secret")
        monkeypatch.setenv("TRELLO_BOARD_ID", "boardid1")
        return TrelloClient.from_env()

    @staticmethod
    def test_init_from_env_vars(client: TrelloClient, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("TRELLO_API_KEY", "api-key2")
        monkeypatch.setenv("TRELLO_API_SECRET", "api-secret2")
        client2 = TrelloClient.from_env()

        assert client is not None
        assert client2 is not None

        assert client._api_key != client2._api_key
        assert client._api_secret != client2._api_secret

    @staticmethod
    def test_list_boards(client: TrelloClient):
        with (Path(__file__).parent / "trello_responses" / "boards.json").open() as f:
            boards_json = json.load(f)

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET, re.compile(r"https://api.trello.com/"), json=boards_json
            )
            boards = client.list_boards()

        assert len(boards) == 1
        assert boards[0].id == "5abbe4b7ddc1b351ef961414"
        assert boards[0].name == "Trello Platform Changes"

    @staticmethod
    def test_get_board(client: TrelloClient):
        with (Path(__file__).parent / "trello_responses" / "board.json").open() as f:
            board_json = json.load(f)

        with (Path(__file__).parent / "trello_responses" / "lists.json").open() as f:
            lists_json = json.load(f)

        with (Path(__file__).parent / "trello_responses" / "cards.json").open() as f:
            cards_json = json.load(f)

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                re.compile(r"https://api.trello.com/1/boards/\w+[^/]+$"),
                json=board_json,
            )
            rsps.add(
                responses.GET,
                re.compile(r"https://api.trello.com/1/boards/\w*/lists"),
                json=lists_json,
            )
            rsps.add(
                responses.GET,
                re.compile(r"https://api.trello.com/1/boards/\w*/cards"),
                json=cards_json,
            )

            board = client.get_board()

        assert board.id == "boardid1"
        assert board.name == "Trello Platform Changes"

        assert len(board.lists) > 0
        a_list = board.lists[0]
        assert a_list.id == "list-id"
        assert a_list.name == "Things to buy today"

        assert len(a_list.cards) > 0
        assert a_list.cards[0].id == "cardid1"
        assert a_list.cards[0].name == "trello CLI with editor interface"

    @staticmethod
    def test_create_card(client: TrelloClient):
        name = "foo bar"
        list_id = "abc123"

        with (Path(__file__).parent / "trello_responses" / "cards.json").open() as f:
            card_json = json.load(f)[0]

        with responses.RequestsMock() as rsps:
            rsps.post(
                re.compile(r"https://api.trello.com/1/cards[^/]+$"),
                match=[
                    responses.json_params_matcher({"name": name, "idList": list_id})
                ],
                json=card_json,
            )
            card_id = client.create_card(name=name, list_id=list_id)

        assert card_id == "cardid1"

    @staticmethod
    def test_change_parent(client: TrelloClient):
        card_id = "cardid1"
        card_json = {
            "id": card_id,
            "name": "trello CLI with editor interface",
            "idList": "oldlistid",
        }
        new_list_id = "newlistid"

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.PUT,
                re.compile(r"https://api.trello.com/1/cards/cardid1[^/]+$"),
                match=[responses.json_params_matcher({"idList": new_list_id})],
                json={**card_json, "idList": new_list_id},
            )

            client.change_parent(card_id=card_id, new_list_id=new_list_id)

    @staticmethod
    def test_rename_card(client: TrelloClient):
        card_id = "cardid1"
        old_card_json = {
            "id": card_id,
            "name": "trello CLI with editor interface",
            "idList": "oldlistid",
        }
        new_name = "new name"

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.PUT,
                re.compile(r"https://api.trello.com/1/cards/cardid1[^/]+$"),
                match=[responses.json_params_matcher({"name": new_name})],
                json={**old_card_json, "name": new_name},
            )

            client.rename_card(card_id=card_id, new_name=new_name)

    @staticmethod
    def test_archive_card(client: TrelloClient):
        card_id = "cardid1"
        old_card_json = {
            "id": card_id,
            "name": "trello CLI with editor interface",
            "idList": "oldlistid",
            "closed": False,
        }

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.PUT,
                re.compile(r"https://api.trello.com/1/cards/cardid1[^/]+$"),
                match=[responses.json_params_matcher({"closed": True})],
                json={**old_card_json, "closed": True},
            )

            client.archive_card(card_id=card_id)



from rebello.models import Board, CardWithID, List

board = Board(
    id="1",
    name="Simple kanban",
    lists=[
        List(
            id="2",
            name="To Do",
            cards=[
                CardWithID(id="21", name="Say hello"),
                CardWithID(id="22", name="Say goodnight"),
            ],
        ),
        List(
            id="3",
            name="Done",
            cards=[
                CardWithID(id="31", name="Basic project setup"),
            ],
        ),
    ],
)



"""
The main models used across the app.

In bigger projects each "module" should depend on its own set of models for modularity, but it's a small project so
a single coupling point is fine.
"""

from dataclasses import dataclass
from typing import Generic, TypeVar


@dataclass
class CardWithID:
    id: str
    name: str


@dataclass
class CardNoID:
    name: str


T = TypeVar("T", covariant=True)


@dataclass
class List(Generic[T]):
    id: str
    name: str
    cards: list[T]


@dataclass
class Board(Generic[T]):
    id: str
    name: str
    lists: list[List[T]]


@dataclass
class ChangeList:
    """Move card from one list to another."""

    card_id: str
    src_list_id: str
    dest_list_id: str


@dataclass
class RewordCard:
    card_id: str
    old_name: str
    new_name: str


@dataclass
class ArchiveCard:
    card_id: str


@dataclass
class CreateCard:
    list_id: str
    name: str


type Change = ChangeList | RewordCard | ArchiveCard | CreateCard
type Changeset = list[Change]



"""
EDI = Editor Driven Interface
"""

from pathlib import Path
import os
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
import subprocess


@contextmanager
def temp_path(ext: str):
    with NamedTemporaryFile(suffix=ext) as file:
        yield Path(file.name)


def open_in_editor(path: Path):
    cmd = [os.environ["EDITOR"], str(path)]
    print(f"Running '{cmd}'")
    proc = subprocess.run(cmd)

    return proc.returncode == 0
